use std::fmt::Debug;

use binrw::{BinRead, BinWrite};
use diff::Diff;

use crate::{
    field_of::FieldOf,
    types::{ObjectType, Coordinate, F64, U32},
};

use super::{ObjectCore, Translate, ObjectModified};

#[cfg_attr(feature = "default-debug", derive(Debug))]
#[derive(BinRead, BinWrite, Clone, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
pub struct Ellipse {
    pub core: ObjectCore,
    #[brw(magic(8u32))] // Number of following fields in struct
    pub clockwise: U32,
    pub corner_a: FieldOf<Coordinate>,
    pub corner_b: FieldOf<Coordinate>,
    pub start_angle: F64, // Radians
    pub end_angle: F64,   // Radians
    pub modified: ObjectModified,
    pub open_curve: U32,
}

// Custom Debug implementation to only print known fields
#[cfg(not(feature = "default-debug"))]
impl Debug for Ellipse {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Ellipse")
            .field("core", &self.core)
            .field("clockwise", &self.clockwise)
            .field("corner_a", &self.corner_a)
            .field("corner_b", &self.corner_b)
            .field("start_angle", &self.start_angle)
            .field("end_angle", &self.end_angle)
            .field("open_curve", &self.open_curve)
            .finish()
    }
}

impl Default for Ellipse {
    fn default() -> Self {
        Self {
            core: ObjectCore::default(ObjectType::Ellipse),
            clockwise: 0.into(),
            corner_a: Coordinate { x: 0.0, y: 0.0 }.into(),
            corner_b: Coordinate { x: 0.0, y: 0.0 }.into(),
            start_angle: 0.0.into(),
            end_angle: 0.0.into(),
            modified: ObjectModified::default(),
            open_curve: 0.into(),
        }
    }
}

impl Translate for Ellipse {
    fn move_absolute(&mut self, origin: Option<Coordinate>, z: Option<f64>) {
        origin.map(|origin| {
            let delta: Coordinate = *self.core.origin - origin;
            *self.corner_a += delta;
            *self.corner_b += delta;
        });
        self.core.move_absolute(origin, z);
    }

    fn move_relative(&mut self, delta: Option<Coordinate>, z: Option<f64>) {
        delta.map(|delta| {
            *self.corner_a += delta;
            *self.corner_b += delta;
        });
        self.core.move_relative(delta, z);
    }
}
